/*
	Copyright (c) 2010 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/

#ifndef SYMMATH_H
#define SYMMATH_H

#include <string>
#include <sstream>

/*
the good news:
automatic differentiation for algebra operations, chain rule, and any functions defined below (most the common math.h ones)

the bad news
using namespace symmath;
cos(2.0)
...will produce a compiler error.
it won't know to pick between ::cos(2.0) and symmath::cos(2.0)
you can get around this by one of two ways:
- either explicitly provide a namespace (above) :  ::cos(2.0) vs. symmath::cos(2.0)
- wrap all your double params in cons() if you want to use the symmath variety but don't want to write "symmath::" before them all
*/

#include <math.h>

namespace symmath {

//everyone wants one.  for differentiation's sake
class var;

/*
TODO
since you can't put doubles as template params...
put a typename for the constant type (always to be cast to an int)
...and then have a special cons class for zero or one (returned with di/dj)
...and then have optimizations for those two special values with the operators
but how would you work around the typedef of the diff(var,var) ? to return either?
you'd have to carry the var ref through template params ... and idk that they do pointers ...
*/
class cons {
public:
	double x;
	cons() : x(0) {}
	cons(double _x) : x(_x) {}
	operator double() const { return x; }

	typedef cons stype;	//source type
	typedef cons dtype;	//deriv type

	//TODO - I should be putting all operators as class methods
	cons operator-() { return cons(-x); }
};
inline cons::dtype diff(const cons &c, const var &x) { return cons(0.0); }
inline std::string debugstr(const cons &c) {
	std::ostringstream o;
	o << c.x;
	return o.str();
}

class var {
public:
	const char *name;
	const double *x;
	var() : x(NULL), name("") {}
	var(const double &x_, const char *name_ = "") : x(&x_), name(name_) {}
	var(const var &src) : x(src.x), name(src.name) {} 
	operator double() const { return *x; }

	typedef var stype;
	typedef cons dtype;

	var &operator=(const var &v) { x = v.x; name = v.name; return *this; }
};
inline var::dtype diff(const var &v, const var &x) { return cons(v.x == x.x); }
inline std::string debugstr(const var &v) {
	std::ostringstream o;
	o << "{";
	if (v.name && *v.name) {
		o << v.name;
	} else {
		o << v.x;
	}
	o << ":" << *v.x << "}"; 
	return o.str();
}

#define VAR(name, ref)	var name(ref, #name)

template<typename T>
class neg {
public:
	T t;
	neg() {}
	neg(const T &_t) : t(_t) {}
	operator double() const { return -(double)t; }
	
	typedef neg<T> stype;
	typedef typename T::dtype dtype;
};
template<typename T> neg<T> inline operator-(const T &t) { return neg<T>(t); }
template<typename T> inline typename neg<T>::dtype diff(const neg<T> &n, const var &x) { return -diff(n.t, x); }
template<typename T> inline std::string debugstr(const neg<T> &n) { return std::string() + "-(" + debugstr(n.t) + ")"; }

template<typename U, typename V>
class add {
public:
	//notice how now we're doing pass-by-value rather than -by-reference
	//this means our var::deriv will be messing up (due to its mem address changing) unless we accomodate
	U u;
	V v;
	add() {}
	add(const U &_u, const V &_v) : u(_u), v(_v) {}
	operator double() const { return (double)u + (double)v; }

	typedef add<U,V> stype;
	typedef add<typename U::dtype, typename V::dtype> dtype;
	
	add &operator=(const add &op) { u = op.u; v = op.v; return *this; }
};
/*
gah, i want this implemented for all my types
but none of the primitive types
how can i do that?
what's better, i want primitives to automatically be cast into const's
*/
template<typename U, typename V> add<U,V> inline operator+(const U &u, const V &v) { return add<U,V>(u,v); }
template<typename V> add<cons,V> inline operator+(double u, const V &v) { return add<cons,V>(cons(u),v); }
template<typename U> add<U,cons> inline operator+(const U &u, double v) { return add<U,cons>(u,cons(v)); }
template<typename U, typename V> inline typename add<U,V>::dtype diff(const add<U,V> &a, const var &x) { return diff(a.u, x) + diff(a.v, x); }
template<typename U, typename V> inline std::string debugstr(const add<U,V> &op) { return std::string() + "(" + debugstr(op.u) + " + " + debugstr(op.v) + ")"; }

template<typename U, typename V>
class sub {
public:
	U u;
	V v;
	sub() {}
	sub(const U &_u, const V &_v) : u(_u), v(_v) {}
	operator double() const { return (double)u - (double)v; }

	typedef sub<U,V> stype;
	typedef sub<typename U::dtype, typename V::dtype> dtype;
};
template<typename U, typename V> sub<U,V> inline operator-(const U &u, const V &v) { return sub<U,V>(u,v); }
template<typename V> sub<cons,V> inline operator-(double u, const V &v) { return sub<cons,V>(cons(u),v); }
template<typename U> sub<U,cons> inline operator-(const U &u, double v) { return sub<U,cons>(u,cons(v)); }
template<typename U, typename V> inline typename sub<U,V>::dtype diff(const sub<U,V> &s, const var &x) { return diff(s.u, x) - diff(s.v, x); }
template<typename U, typename V> inline std::string debugstr(const sub<U,V> &op) { return std::string() + "(" + debugstr(op.u) + " - " + debugstr(op.v) + ")"; }

template<typename U, typename V>
class mul {
public:
	U u;
	V v;
	mul() {}
	mul(const U &_u, const V &_v) : u(_u), v(_v) {}
	operator double() const { return (double)u * (double)v; }

	typedef mul<U,V> stype;
	typedef add<mul<typename U::dtype, V>, mul<U, typename V::dtype> > dtype;
};
template<typename U, typename V> mul<U,V> inline operator*(const U &u, const V &v) { return mul<U,V>(u,v); }
template<typename V> mul<cons,V> inline operator*(double u, const V &v) { return mul<cons,V>(cons(u),v); }
template<typename U> mul<U,cons> inline operator*(const U &u, double v) { return mul<U,cons>(u,cons(v)); }
template<typename U, typename V> inline typename mul<U,V>::dtype diff(const mul<U,V> &m, const var &x) { return diff(m.u, x) * m.v + m.u * diff(m.v, x); }
template<typename U, typename V> inline std::string debugstr(const mul<U,V> &op) { return std::string() + "(" + debugstr(op.u) + " * " + debugstr(op.v) + ")"; }

template<typename U, typename V>
class div {
public:
	U u;
	V v;
	div() {}
	div(const U &_u, const V &_v) : u(_u), v(_v) {}
	operator double() const { return (double)u / (double)v; }
	
	typedef div<U,V> stype;
	typedef div<sub<mul<typename U::dtype, V>, mul<U, typename V::dtype> >, mul<V,V> > dtype;
};
template<typename U, typename V> div<U,V> inline operator/(const U &u, const V &v) { return div<U,V>(u,v); }
template<typename V> div<cons,V> inline operator/(double u, const V &v) { return div<cons,V>(cons(u),v); }
template<typename U> div<U,cons> inline operator/(const U &u, double v) { return div<U,cons>(u,cons(v)); }
template<typename U, typename V> inline typename div<U,V>::dtype diff(const div<U,V> &d, const var &x) { return (diff(d.u, x) * d.v - d.u * diff(d.v, x)) / (d.v * d.v); }
template<typename U, typename V> inline std::string debugstr(const div<U,V> &op) { return std::string() + "(" + debugstr(op.u) + " / " + debugstr(op.v) + ")"; }


#define DEFINE_FUNC_REAL_TO_REAL(name, calc, derivtype)	\
template<typename T>	\
class name##_op {	\
public:	\
	T t;	\
	name##_op() {}	\
	name##_op(const T &t_) : t(t_) {}	\
	operator double() const { double t = (double)this->t; return (double)(calc); }	\
	typedef name##_op<T> stype;	\
	typedef mul< derivtype, typename T::dtype> dtype;	\
};	\
template<typename T> inline name##_op<T> name(const T &t) { return name##_op<T>(t); }	\
inline name##_op<cons> name(double t) { return name##_op<cons>(cons(t)); }	\
template<typename T> inline typename name##_op<T>::dtype diff(const name##_op<T> &f, const var &x) { const T &t = f.t; return diff(calc, t) * diff(t, x); }	\
template<typename T> inline std::string debugstr(const name##_op<T> &f) { return std::string() + #name "(" + debugstr(f.t) + ")"; }


//first is the name/binding, second is the return type, third is the stmt type, fourth is the stmt (in symbols)
#define DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(binding,name,derivtype,derivcalc)	\
template<typename T>	\
class name##_op {	\
public:	\
	T t;	\
	name##_op() {}	\
	name##_op(const T &_t) : t(_t) {}	\
	operator double() const { return binding((double)t); }	\
	typedef name##_op<T> stype;	\
	typedef mul< derivtype, typename T::dtype> dtype;	\
};	\
template<typename T> inline name##_op<T> name(const T &t) { return name##_op<T>(t); }	\
name##_op<cons> inline name(double t) { return name##_op<cons>(cons(t)); }	\
template<typename T> inline typename name##_op<T>::dtype diff(const name##_op<T> &f, const var &x) { const T &t = f.t; return (derivcalc) * diff(t, x); }	\
template<typename T> inline std::string debugstr(const name##_op<T> &f) { return std::string() + #name "(" + debugstr(f.t) + ")"; }

//to get past mixing template param lists and macro param lists...
#define COMMA	,

#if 0	//new method - less wrappers to build functions of other functions.  hopefully i'll get debugstr working with it too... it's also failing on chainrule...
//todo - for debug string's simplicity's sake - make this a class of its own
DEFINE_FUNC_REAL_TO_REAL(square, t * t, typename mul<T COMMA T>::dtype)
#else	//old method, requires a lot of explicit stuff provided...
//this is used often enough, it should be a function for shortcut's sake
inline double square_func(double x) { return x * x; }
DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(square_func, square, mul<cons COMMA T>, 2. * t); //d/dx[x*x] = 2*x
//TODO - same for 1/x
//and then see if you can write in optimizations for special-case templates (to weed out times-zero and the likes)
#endif

DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(::exp, exp, exp_op<T>, exp(t))			//d/dx[exp(x)] = exp(x)

DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(::log, log, div<cons COMMA T>, 1. / t)						//d/dx[log(x)] = 1/x
DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(::sqrt, sqrt, div<cons COMMA sqrt_op<T> >, .5 / sqrt(t))	//d/dx[sqrt(x)] = d/dx[x^.5] = d/dx[.5 * x^-.5] = d/dx[ .5/sqrt(x) ]

//d/dx[ u^v ] = d/dx[ exp(log(u^v)) ] = d/dx[ exp(v*log(u)) ] = exp(v*log(u)) * d/dx[v*log(u)] = u^v * (dv/dx * log(u) + v * d/dx[log(u)]) = u^v * (dv/dx * log(u) + v / u * du/dx)
//... i'll just rewrite it in terms of exp and log
template<typename U, typename V> inline exp_op<mul<V, log_op<U> > > pow(U &u, V &v) { return exp(v * log(u)); }
//these two can be optimized...
template<typename V> inline exp_op<mul<V, log_op<cons> > > pow(double u, V &v) { return exp(v * log(cons(u))); }
template<typename U> inline exp_op<mul<cons, log_op<U> > > pow(U &u, double v) { return exp(cons(v) * log(u)); }


// trigonometric functions


//here's our circular def of derivs of cos and sin
//suffix '_op' is appended to the function name to create the hidden classes
template<typename T> class cos_op;
template<typename T> class sin_op;

DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(::cos, cos, neg<sin_op<T> >, -sin(t))		//d/dx[cos(x)] = -sin(x)
DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(::sin, sin, cos_op<T>, cos(t))			//d/dx[sin(x)] = cos(x)
DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(::tan, tan, add<cons COMMA square_op<tan_op<T> > >, 1. + square(tan(t)))


// inverse trigonometric functions


DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(::acos, acos, div<cons COMMA sqrt_op<sub<cons COMMA square_op<T> > > >, -1. / sqrt(1. - square(t)))	//d/dx[acos(x)] = -1/sqrt(1-x^2)
DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(::asin, asin, div<cons COMMA sqrt_op<sub<cons COMMA square_op<T> > > >, 1. / sqrt(1. - square(t)))	//d/dx[asin(x)] = 1/sqrt(1-x^2)
DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(::atan, atan, div<cons COMMA add<cons COMMA square_op<T> > >, 1. / (1. + square(t)))					//d/dx[atan(x)] = 1/(1+x*x)


// common ansi math functions


inline double sgn_func(double x) { return x > 0. ? 1. : -1.; }
DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(sgn_func, sgn, cons, 0.)	//d/dx[sgn(x)] = {x=0: undefined, else: 0}
DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(::fabs, fabs, sgn_op<T>, sgn(t))	//d/dx[|x|] = sgn(x)
//DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(::fmod, fmod, cons, 1.)	//d/dx[fmod(u,v)] = ... hmm ... d/dx[fmod(x,c)] = 1 ... but what about a varying 2nd param?


// hyperbolic functions


//TODO - make a class like the one above except that auto-diffs itself (rather than relies on a diff to be provided)
//...then you wouldn't have to do the extensive type substitution as we see in tanh...
#define COSHTYPE mul<cons, add<exp_op<T>, exp_op<neg<T> > > >
template<typename T> inline COSHTYPE cosh(const T &t) { return .5 * (exp(t) + exp(-t)); }
#define SINHTYPE mul<cons, sub<exp_op<T>, exp_op<neg<T> > > >
template<typename T> inline SINHTYPE sinh(const T &t) { return .5 * (exp(t) - exp(-t)); }
template<typename T> inline div<SINHTYPE, COSHTYPE > tanh(const T &t) { return sinh(t) / cosh(t); }
#undef COSHTYPE
#undef SINHTYPE


// inverse hyperoblic functions.  NOTICE - may not be in all math.h's


template<typename X> inline log_op<add<X, sqrt_op<add<square_op<X>, cons> > > > asinh(const X &x) { return log(x + sqrt(square(x) + 1.)); }
template<typename X> inline log_op<add<X, sqrt_op<sub<square_op<X>, cons> > > > acosh(const X &x) { return log(x + sqrt(square(x) - 1.)); }
template<typename X> inline mul<cons, log_op<div<add<cons, X>, sub<cons, X> > > > atanh(const X &x) { return .5 * log((1. + x) / (1. - x)); }


// extra functions


DEFINE_FUNC_REAL_TO_REAL_WITH_DERIV(::cbrt, cbrt, div<cons COMMA mul<cons COMMA cbrt_op<square_op<T> > > >, 1./(3. * cbrt(square(t)))) 	//d/dx[x^(1/3)] = 1/3 x^(-2/3) = 1/(3 * (x^2)^(1/3) )
//erf
//erfc
//lgamma
//tgamma

};

#endif	//SYMMATH_H
